/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "../client/client.h"
#include "../client/qmenu.h"

extern cvar_t *vid_ref;
extern cvar_t *vid_fullscreen;
extern cvar_t *vid_gamma;
extern cvar_t *scr_viewsize;

// Knightmare- cvars for advanced GL options menu
cvar_t	*gl_modulate;
cvar_t	*intensity;
cvar_t	*r_overbrightbits;
cvar_t	*gl_trans_lightmaps;
cvar_t	*gl_pixel_shader_warp;
cvar_t	*r_waterwave;
cvar_t	*gl_shadows;
cvar_t	*gl_envmaps;
cvar_t	*gl_glass_envmaps;
cvar_t	*gl_screenshot_jpeg;
cvar_t	*gl_screenshot_jpeg_quality;
// end Knightmare

static cvar_t *gl_mode;
static cvar_t *gl_driver;
static cvar_t *gl_picmip;
static cvar_t *gl_texturemode;
static cvar_t *gl_anisotropic;
static cvar_t *gl_finish;
static cvar_t *con_MENU_FONT_SIZE;


extern void M_ForceMenuOff( void );
void M_PushMenu ( void (*draw) (void), const char *(*key) (int k) );

/*
====================================================================

MENU INTERACTION

====================================================================
*/

#define OPENGL_MENU   0
#define OPENGL_ADVANCED_MENU   1

static menuframework_s	s_opengl_menu;
static menuframework_s	s_opengl_menu_advanced; // Knightmare added
static menuframework_s *s_current_menu;

static menulist_s		s_mode_list;
static menulist_s  		s_fs_box;
static menuslider_s		s_screensize_slider;
static menuslider_s		s_brightness_slider;
static menulist_s		s_texqual_box;
static menulist_s		s_tfilter_box;
static menulist_s  		s_texcompress_box;
static menulist_s  		s_finish_box;
static menuaction_s		s_advanced_action; // Knightmare added
//Knightmare- replace cancel with apply, thnx to MrG
static menuaction_s		s_defaults_action;
static menuaction_s		s_apply_action[2];


// Knightmare- items for advanced GL options menu
static menuseparator_s	s_options_advanced_header;	
static menuslider_s		s_lightmapscale_slider;
static menuslider_s		s_textureintensity_slider;
static menulist_s  		s_overbrights_box;
static menulist_s  		s_trans_lightmap_box;
static menulist_s  		s_texshader_warp_box;
static menuslider_s  	s_waterwave_slider;
static menulist_s		s_glshadows_box;
static menulist_s  		s_ent_envmap_box;
static menulist_s  		s_glass_envmap_box;
static menulist_s  		s_screenshotjpeg_box;
static menuslider_s  	s_screenshotjpegquality_slider;
static menuaction_s		s_back_action;
// end Knightmare

// Knightmare- functions for advanced GL options menu
void VID_MenuAdvancedInit (void);
void VID_MenuAdvancedDraw (void);

void M_Menu_Video_Advanced_f (void)
{
	VID_MenuAdvancedInit();
	M_PushMenu( VID_MenuAdvancedDraw, VID_MenuKey );
}

static void AdvancedOptions( void *s )
{
	M_Menu_Video_Advanced_f ();
}

static float ClampCvar( float min, float max, float value )
{
	if ( value < min ) return min;
	if ( value > max ) return max;
	return value;
}

static void VID_MenuAdvancedSetValues ( void )
{
	Cvar_SetValue( "gl_modulate", ClampCvar( 1, 2, gl_modulate->value ) );
	s_lightmapscale_slider.curvalue = (Cvar_VariableValue("gl_modulate") -1) * 10;

	Cvar_SetValue( "intensity", ClampCvar( 1, 2, intensity->value ) );
	s_textureintensity_slider.curvalue = (Cvar_VariableValue("intensity") -1) * 10;

	Cvar_SetValue( "r_overbrightbits", ClampCvar( 1, 2, r_overbrightbits->value ) );
	if (r_overbrightbits->value == 1)
		s_overbrights_box.curvalue = 0;
	else
		s_overbrights_box.curvalue = 1;

	Cvar_SetValue( "gl_trans_lightmaps", ClampCvar( 0, 1, gl_trans_lightmaps->value ) );
	s_trans_lightmap_box.curvalue	= Cvar_VariableValue("gl_trans_lightmaps");

	Cvar_SetValue( "gl_pixel_shader_warp", ClampCvar( 0, 1, gl_pixel_shader_warp->value ) );
	s_texshader_warp_box.curvalue	= Cvar_VariableValue("gl_pixel_shader_warp");

	s_waterwave_slider.curvalue		= Cvar_VariableValue("r_waterwave");

	Cvar_SetValue( "gl_shadows", ClampCvar( 0, 2, gl_shadows->value ) );
	s_glshadows_box.curvalue	= Cvar_VariableValue("gl_shadows");

	Cvar_SetValue( "gl_envmaps", ClampCvar( 0, 1, gl_envmaps->value ) );
	s_ent_envmap_box.curvalue	= Cvar_VariableValue("gl_envmaps");

	Cvar_SetValue( "gl_glass_envmaps", ClampCvar( 0, 1, gl_glass_envmaps->value ) );
	s_glass_envmap_box.curvalue	= Cvar_VariableValue("gl_glass_envmaps");

	Cvar_SetValue( "gl_screenshot_jpeg", ClampCvar( 0, 1, gl_screenshot_jpeg->value ) );
	s_screenshotjpeg_box.curvalue	= Cvar_VariableValue("gl_screenshot_jpeg");

	Cvar_SetValue( "gl_screenshot_jpeg_quality", ClampCvar( 50, 100, gl_screenshot_jpeg_quality->value ) );
	s_screenshotjpegquality_slider.curvalue	= (Cvar_VariableValue("gl_screenshot_jpeg_quality") -50) / 5;
}

static void LightMapScaleCallback ( void *unused )
{
	Cvar_SetValue( "gl_modulate", s_lightmapscale_slider.curvalue / 10 + 1);
}

static void TextureIntensCallback ( void *unused )
{
	Cvar_SetValue( "intensity", s_textureintensity_slider.curvalue / 10 + 1);
}

static void OverbrightsCallback ( void *unused )
{
	Cvar_SetValue( "r_overbrightbits", s_overbrights_box.curvalue + 1);
}

static void TransFullbrightCallback ( void *unused )
{
	Cvar_SetValue( "gl_trans_lightmaps", s_trans_lightmap_box.curvalue);
}

static void TexShaderWarpCallback ( void *unused )
{
	Cvar_SetValue( "gl_pixel_shader_warp", s_texshader_warp_box.curvalue);
}

static void WaterWaveCallback ( void *unused )
{
	Cvar_SetValue( "r_waterwave", s_waterwave_slider.curvalue);
}

static void GLShadowsCallback ( void *unused )
{
	Cvar_SetValue( "gl_shadows", s_glshadows_box.curvalue);
}

static void EntEnvmapCallback ( void *unused )
{
	Cvar_SetValue( "gl_envmaps", s_ent_envmap_box.curvalue);
}

static void GlassEnvmapCallback ( void *unused )
{
	Cvar_SetValue( "gl_glass_envmaps", s_glass_envmap_box.curvalue);
}

static void JPEGScreenshotCallback ( void *unused )
{
	Cvar_SetValue( "gl_screenshot_jpeg", s_screenshotjpeg_box.curvalue);
}

static void JPEGScreenshotQualityCallback ( void *unused )
{
	Cvar_SetValue( "gl_screenshot_jpeg_quality", (s_screenshotjpegquality_slider.curvalue * 5 + 50));
}
// end Knightmare



static void ScreenSizeCallback( void *s )
{
	menuslider_s *slider = ( menuslider_s * ) s;

	Cvar_SetValue( "viewsize", slider->curvalue * 10 );
}

static void BrightnessCallback( void *s )
{
	menuslider_s *slider = ( menuslider_s * ) s;
	// invert sense so greater = brighter, and scale to a range of 0.5 to 1.3
	Cvar_SetValue( "vid_gamma", ( 0.8 - ( s_brightness_slider.curvalue/10.0 - 0.5 ) ) + 0.5 );
}

static void ResetDefaults( void *unused )
{
	VID_MenuInit();
}

static void prepareVideoRefresh( void )
{
	//set the right mode for refresh
	Cvar_Set( "vid_ref", "kmgl" );
	Cvar_Set( "gl_driver", "opengl32" );

	//tell them theyr're modified so they refresh
	if ( vid_gamma->modified )
		vid_ref->modified = true;
	if ( gl_driver->modified )
		vid_ref->modified = true;
}

static void ApplyChanges( void *unused )
{
	if (s_tfilter_box.curvalue == 0)
		Cvar_Set("gl_texturemode", "GL_NEAREST");
	else if (s_tfilter_box.curvalue == 1)
		Cvar_Set("gl_texturemode", "GL_NEAREST_MIPMAP_NEAREST");
	else if (s_tfilter_box.curvalue == 2)
		Cvar_Set("gl_texturemode", "GL_LINEAR");
	else if (s_tfilter_box.curvalue == 3)
		Cvar_Set("gl_texturemode", "GL_LINEAR_MIPMAP_NEAREST");
	else if (s_tfilter_box.curvalue == 4)
		Cvar_Set("gl_texturemode", "GL_LINEAR_MIPMAP_LINEAR");
	else if (s_tfilter_box.curvalue == 5)
		Cvar_Set("gl_texturemode", "GL_LINEAR_MIPMAP_NEAREST");
	else //if (s_tfilter_box.curvalue == 6)
		Cvar_Set("gl_texturemode", "GL_LINEAR_MIPMAP_LINEAR");

	Cvar_SetValue( "gl_mode", s_mode_list.curvalue + 3 ); // offset for eliminating < 640x480 modes
	Cvar_SetValue( "vid_fullscreen", s_fs_box.curvalue );
	// invert sense so greater = brighter, and scale to a range of 0.5 to 1.3
	Cvar_SetValue( "vid_gamma", ( 0.8 - ( s_brightness_slider.curvalue/10.0 - 0.5 ) ) + 0.5 );
	Cvar_SetValue( "gl_picmip", 2-s_texqual_box.curvalue );
	Cvar_SetValue( "gl_anisotropic", (s_tfilter_box.curvalue > 4) );
	Cvar_SetValue( "gl_ext_texture_compression", s_texcompress_box.curvalue );
	Cvar_SetValue( "gl_finish", s_finish_box.curvalue );

	prepareVideoRefresh ();

	M_ForceMenuOff();
}

static void AdvancedMenuApplyChanges ( void *unused )
{
//	extern void M_PopMenu( void );
	// update for modified intensity
	if ( intensity->modified )
		vid_ref->modified = true;
//	M_PopMenu();
}

static void CancelChanges( void *unused )
{
	extern void M_PopMenu( void );

	M_PopMenu();
}

/*
** VID_MenuInit
*/

// Knightmare- init advanced GL options menu
void VID_MenuAdvancedInit (void)
{
	static const char *yesno_names[] =
	{
		"no",
		"yes",
		0
	};
	static const char *shadow_names[] =
	{
		"no",
		"static",
		"dynamic",
		0
	};
	static const char *ifsupported_names[] =
	{
		"no",
		"if supported",
		0
	};
	int y = 0;

	gl_modulate = Cvar_Get( "gl_modulate", "1", CVAR_ARCHIVE );
	intensity = Cvar_Get ("intensity", "1", 0);
	r_overbrightbits = Cvar_Get ("r_overbrightbits", "2", CVAR_ARCHIVE);
	gl_trans_lightmaps = Cvar_Get( "gl_trans_lightmaps", "1", CVAR_ARCHIVE );
	gl_pixel_shader_warp = Cvar_Get( "gl_pixel_shader_warp", "1", CVAR_ARCHIVE );
	r_waterwave = Cvar_Get ("r_waterwave", "0", CVAR_ARCHIVE );
	gl_shadows = Cvar_Get ("gl_shadows", "0", CVAR_ARCHIVE );
	gl_envmaps = Cvar_Get( "gl_envmaps", "1", CVAR_ARCHIVE );
	gl_glass_envmaps = Cvar_Get( "gl_glass_envmaps", "1", CVAR_ARCHIVE );
	gl_screenshot_jpeg = Cvar_Get( "gl_screenshot_jpeg", "1", CVAR_ARCHIVE );
	gl_screenshot_jpeg_quality = Cvar_Get( "gl_screenshot_jpeg_quality", "85", CVAR_ARCHIVE );	// Heffo - JPEG Screenshots


	s_opengl_menu_advanced.x = viddef.width * 0.50;
	s_opengl_menu_advanced.nitems = 0;

	s_options_advanced_header.generic.type	= MTYPE_SEPARATOR;
	s_options_advanced_header.generic.name	= "Advanced Options";
	s_options_advanced_header.generic.x	= MENU_FONT_SIZE/2 * strlen(s_options_advanced_header.generic.name);
	s_options_advanced_header.generic.y	= 0;

	s_lightmapscale_slider.generic.type	= MTYPE_SLIDER;
	s_lightmapscale_slider.generic.x		= 0;
	s_lightmapscale_slider.generic.y		= y += 3*(MENU_FONT_SIZE+2); //10
	s_lightmapscale_slider.generic.name	= "lightmap scale";
	s_lightmapscale_slider.generic.callback = LightMapScaleCallback;
	s_lightmapscale_slider.minvalue = 0;
	s_lightmapscale_slider.maxvalue = 10;

	s_textureintensity_slider.generic.type	= MTYPE_SLIDER;
	s_textureintensity_slider.generic.x		= 0;
	s_textureintensity_slider.generic.y		= y += (MENU_FONT_SIZE+2); //20
	s_textureintensity_slider.generic.name	= "texture intensity";
	s_textureintensity_slider.generic.callback = TextureIntensCallback;
	s_textureintensity_slider.minvalue = 0;
	s_textureintensity_slider.maxvalue = 10;

	s_overbrights_box.generic.type = MTYPE_SPINCONTROL;
	s_overbrights_box.generic.x	= 0;
	s_overbrights_box.generic.y	= y += (MENU_FONT_SIZE+2); //30
	s_overbrights_box.generic.name	= "overbrights";
	s_overbrights_box.generic.callback = OverbrightsCallback;
	s_overbrights_box.itemnames = yesno_names;

	s_trans_lightmap_box.generic.type		= MTYPE_SPINCONTROL;
	s_trans_lightmap_box.generic.x		= 0;
	s_trans_lightmap_box.generic.y		= y += (MENU_FONT_SIZE+2); //40
	s_trans_lightmap_box.generic.name		= "translucent lightmaps";
	s_trans_lightmap_box.generic.callback	= TransFullbrightCallback;
	s_trans_lightmap_box.itemnames		= yesno_names;

	s_texshader_warp_box.generic.type		= MTYPE_SPINCONTROL;
	s_texshader_warp_box.generic.x			= 0;
	s_texshader_warp_box.generic.y			= y += (MENU_FONT_SIZE+2); //50
	s_texshader_warp_box.generic.name		= "texture shader warp";
	s_texshader_warp_box.generic.callback	= TexShaderWarpCallback;
	s_texshader_warp_box.itemnames			= ifsupported_names;

	s_waterwave_slider.generic.type		= MTYPE_SLIDER;
	s_waterwave_slider.generic.x		= 0;
	s_waterwave_slider.generic.y		= y += (MENU_FONT_SIZE+2); //60
	s_waterwave_slider.generic.name		= "water wave size";
	s_waterwave_slider.generic.callback	= WaterWaveCallback;
	s_waterwave_slider.minvalue			= 0;
	s_waterwave_slider.maxvalue			= 24;

	s_glshadows_box.generic.type		= MTYPE_SPINCONTROL;
	s_glshadows_box.generic.x			= 0;
	s_glshadows_box.generic.y			= y += (MENU_FONT_SIZE+2); //70
	s_glshadows_box.generic.name		= "entity shadows";
	s_glshadows_box.generic.callback	= GLShadowsCallback;
	s_glshadows_box.itemnames			= shadow_names;

	s_ent_envmap_box.generic.type		= MTYPE_SPINCONTROL;
	s_ent_envmap_box.generic.x			= 0;
	s_ent_envmap_box.generic.y			= y += (MENU_FONT_SIZE+2); //80
	s_ent_envmap_box.generic.name		= "entity envmaps";
	s_ent_envmap_box.generic.callback	= EntEnvmapCallback;
	s_ent_envmap_box.itemnames			= yesno_names;

	s_glass_envmap_box.generic.type		= MTYPE_SPINCONTROL;
	s_glass_envmap_box.generic.x			= 0;
	s_glass_envmap_box.generic.y			= y += (MENU_FONT_SIZE+2); //90
	s_glass_envmap_box.generic.name		= "glass envmaps";
	s_glass_envmap_box.generic.callback	= GlassEnvmapCallback;
	s_glass_envmap_box.itemnames			= yesno_names;

	s_screenshotjpeg_box.generic.type		= MTYPE_SPINCONTROL;
	s_screenshotjpeg_box.generic.x			= 0;
	s_screenshotjpeg_box.generic.y			= y += (MENU_FONT_SIZE+2); //100
	s_screenshotjpeg_box.generic.name		= "JPEG screenshots";
	s_screenshotjpeg_box.generic.callback	= JPEGScreenshotCallback;
	s_screenshotjpeg_box.itemnames			= yesno_names;

	s_screenshotjpegquality_slider.generic.type	= MTYPE_SLIDER;
	s_screenshotjpegquality_slider.generic.x	= 0;
	s_screenshotjpegquality_slider.generic.y	= y += (MENU_FONT_SIZE+2); //110
	s_screenshotjpegquality_slider.generic.name	= "JPEG screenshot quality";
	s_screenshotjpegquality_slider.generic.callback		= JPEGScreenshotQualityCallback;
	s_screenshotjpegquality_slider.minvalue		= 0;
	s_screenshotjpegquality_slider.maxvalue		= 10;

	s_apply_action[OPENGL_ADVANCED_MENU].generic.type = MTYPE_ACTION;
	s_apply_action[OPENGL_ADVANCED_MENU].generic.name = "apply changes";
	s_apply_action[OPENGL_ADVANCED_MENU].generic.x    = 0;
	s_apply_action[OPENGL_ADVANCED_MENU].generic.y    = y += 2*(MENU_FONT_SIZE+2); //130
	s_apply_action[OPENGL_ADVANCED_MENU].generic.callback = AdvancedMenuApplyChanges;

	s_back_action.generic.type = MTYPE_ACTION;
	s_back_action.generic.name = "back";
	s_back_action.generic.x    = 0;
	s_back_action.generic.y    = y += (MENU_FONT_SIZE+2); //140
	s_back_action.generic.callback = CancelChanges;

	VID_MenuAdvancedSetValues();

	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_options_advanced_header );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_lightmapscale_slider );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_textureintensity_slider );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_overbrights_box );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_trans_lightmap_box );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_texshader_warp_box );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_waterwave_slider );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_glshadows_box );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_ent_envmap_box );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_glass_envmap_box );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_screenshotjpeg_box );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_screenshotjpegquality_slider );
	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_apply_action[OPENGL_ADVANCED_MENU] );

	Menu_AddItem( &s_opengl_menu_advanced, ( void * ) &s_back_action );

	Menu_Center( &s_opengl_menu_advanced );

	s_opengl_menu_advanced.x -= MENU_FONT_SIZE;	
}
// end Knightmare

// Knightmare added
int texfilter_box_setval (void)
{
	if (!Q_strcasecmp(gl_texturemode->string, "GL_NEAREST"))
		return 0;
	else if (!Q_strcasecmp(gl_texturemode->string, "GL_NEAREST_MIPMAP_NEAREST"))
		return 1;
	else if (!Q_strcasecmp(gl_texturemode->string, "GL_LINEAR"))
		return 2;
	else if (!Q_strcasecmp(gl_texturemode->string, "GL_LINEAR_MIPMAP_NEAREST"))
		return (gl_anisotropic->value)? 5 :3;
	else // if (!Q_strcasecmp(gl_texturemode->string, "GL_LINEAR_MIPMAP_LINEAR"))
		return (gl_anisotropic->value)? 6 :4;
}

void VID_MenuInit (void)
{
	// Knightmare- added 1280x1024, 1400x1050, 856x480, 1024x480 modes, removed 320x240, 400x300, 512x384 modes
	static const char *resolutions[] = 
	{
		"[640 480  ]",
		"[800 600  ]",
		"[960 720  ]",
		"[1024 768 ]",
		"[1152 864 ]",
		"[1280 960 ]",
		"[1280 1024]", // Knightmare added
		"[1400 1050]", // Knightmare added
		"[1600 1200]",
		"[2048 1536]",
		"[856 480  ]", // Knightmare added
		"[1024 480 ]", // Knightmare added
		0
	};
	static const char *yesno_names[] =
	{
		"no",
		"yes",
		0
	};
	static const char *mip_names[] =
	{
		"none",
		"nearest",
		"linear",
		"bilinear",
		"trilinear",
		"bilinear anisotropic",
		"trilinear anisotopic",		
		0
	};
	static const char *lmh_names[] =
	{
		"low",
		"medium",
		"high",		
		0
	};

	int  y;

	if ( !gl_driver )
		gl_driver = Cvar_Get( "gl_driver", "opengl32", 0 );
	if ( !gl_picmip )
		gl_picmip = Cvar_Get( "gl_picmip", "0", 0 );
	if ( !gl_mode )
		gl_mode = Cvar_Get( "gl_mode", "3", 0 ); // default to 640x480
	if ( !gl_finish )
		gl_finish = Cvar_Get( "gl_finish", "0", CVAR_ARCHIVE );
	if ( !gl_texturemode )
		gl_texturemode = Cvar_Get( "gl_texturemode", "GL_LINEAR_MIPMAP_NEAREST", CVAR_ARCHIVE );
	if ( !gl_anisotropic )
		gl_anisotropic = Cvar_Get( "gl_anisotropic", "0", CVAR_ARCHIVE );


	s_mode_list.curvalue = gl_mode->value - 3; // offset for getting rid of < 640x480 resolutions
	if (s_mode_list.curvalue < 0) s_mode_list.curvalue = 0; // clamp to minimum

	if ( !scr_viewsize )
		scr_viewsize = Cvar_Get ("viewsize", "100", CVAR_ARCHIVE);

	if ( !con_font_size )
		con_font_size = Cvar_Get ("con_font_size", "8", CVAR_ARCHIVE);

	s_screensize_slider.curvalue = scr_viewsize->value/10;

	s_opengl_menu.x = viddef.width * 0.50;
	s_opengl_menu.nitems = 0;


	y = 0;
	s_mode_list.generic.type = MTYPE_SPINCONTROL;
	s_mode_list.generic.name = "video mode";
	s_mode_list.generic.x = 0;
	s_mode_list.generic.y = y;
	s_mode_list.itemnames = resolutions;

	s_fs_box.generic.type = MTYPE_SPINCONTROL;
	s_fs_box.generic.x	= 0;
	s_fs_box.generic.y	= y += (MENU_FONT_SIZE+2); //10
	s_fs_box.generic.name	= "fullscreen";
	s_fs_box.itemnames = yesno_names;
	s_fs_box.curvalue = vid_fullscreen->value;

	s_screensize_slider.generic.type	= MTYPE_SLIDER;
	s_screensize_slider.generic.x		= 0;
	s_screensize_slider.generic.y		= y += (MENU_FONT_SIZE+2); //20
	s_screensize_slider.generic.name	= "screen size";
	s_screensize_slider.minvalue = 3;
	s_screensize_slider.maxvalue = 12;
	s_screensize_slider.generic.callback = ScreenSizeCallback;

	s_brightness_slider.generic.type	= MTYPE_SLIDER;
	s_brightness_slider.generic.x	= 0;
	s_brightness_slider.generic.y	= y += (MENU_FONT_SIZE+2); //30
	s_brightness_slider.generic.name	= "brightness";
	s_brightness_slider.generic.callback = BrightnessCallback;
	s_brightness_slider.minvalue = 5;
	s_brightness_slider.maxvalue = 13;
	s_brightness_slider.curvalue = ( 1.3 - vid_gamma->value + 0.5 ) * 10;

	s_tfilter_box.generic.type = MTYPE_SPINCONTROL;
	s_tfilter_box.generic.x	= 0;
	s_tfilter_box.generic.y	= y += (MENU_FONT_SIZE+2); //40
	s_tfilter_box.generic.name	= "texture filter";
	s_tfilter_box.itemnames = mip_names;
	s_tfilter_box.curvalue = texfilter_box_setval();

	s_texqual_box.generic.type	= MTYPE_SPINCONTROL;
	s_texqual_box.generic.x		= 0;
	s_texqual_box.generic.y		= y += (MENU_FONT_SIZE+2); //50
	s_texqual_box.generic.name	= "texture quality";
	s_texqual_box.itemnames = lmh_names;
	s_texqual_box.curvalue = ClampCvar (0, 2, 2-gl_picmip->value);

	s_texcompress_box.generic.type		= MTYPE_SPINCONTROL;
	s_texcompress_box.generic.x			= 0;
	s_texcompress_box.generic.y			= y += (MENU_FONT_SIZE+2); //60
	s_texcompress_box.generic.name		= "texture compression";
	s_texcompress_box.itemnames			= yesno_names;
	s_texcompress_box.curvalue			= Cvar_VariableValue("gl_ext_texture_compression");

	s_finish_box.generic.type = MTYPE_SPINCONTROL;
	s_finish_box.generic.x	= 0;
	s_finish_box.generic.y	= y += 2*(MENU_FONT_SIZE+2); //80
	s_finish_box.generic.name	= "sync every frame";
	s_finish_box.curvalue = gl_finish->value;
	s_finish_box.itemnames = yesno_names;

	// Knightmare added
	s_advanced_action.generic.type = MTYPE_ACTION;
	s_advanced_action.generic.name = "advanced options";
	s_advanced_action.generic.x    = 0;
	s_advanced_action.generic.y    = y += 2*(MENU_FONT_SIZE+2); //100
	s_advanced_action.generic.callback = AdvancedOptions;

	s_defaults_action.generic.type = MTYPE_ACTION;
	s_defaults_action.generic.name = "reset to defaults";
	s_defaults_action.generic.x    = 0;
	s_defaults_action.generic.y    = y += 2*(MENU_FONT_SIZE+2); //120
	s_defaults_action.generic.callback = ResetDefaults;

	//Knightmare- changed cancel to apply changes, thanx to MrG
	s_apply_action[OPENGL_MENU].generic.type = MTYPE_ACTION;
	s_apply_action[OPENGL_MENU].generic.name = "apply changes";
	s_apply_action[OPENGL_MENU].generic.x    = 0;
	s_apply_action[OPENGL_MENU].generic.y    = y += (MENU_FONT_SIZE+2); //130
	s_apply_action[OPENGL_MENU].generic.callback = ApplyChanges;
	//end Knightmare

	Menu_AddItem( &s_opengl_menu, ( void * ) &s_mode_list );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_fs_box );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_screensize_slider );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_brightness_slider );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_tfilter_box );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_texqual_box );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_texcompress_box );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_finish_box );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_advanced_action ); // Knightmare added

	Menu_AddItem( &s_opengl_menu, ( void * ) &s_defaults_action );
	//Knightmare- changed cancel to apply changes, thanx to MrG
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_apply_action[OPENGL_MENU] );

	Menu_Center( &s_opengl_menu );
	s_opengl_menu.x -= MENU_FONT_SIZE;
}

/*
================
VID_MenuDraw
================
*/
void M_Banner( char *name );

void VID_MenuDraw (void)
{
	//int w, h;

	//if ( s_current_menu_index == 0 )
	//	s_current_menu = &s_software_menu;
	//else
		s_current_menu = &s_opengl_menu;

	/*
	** draw the banner
	*/
	//re.DrawGetPicSize( &w, &h, "m_banner_video" );
	//re.DrawPic( viddef.width / 2 - w / 2, viddef.height /2 - 110, "m_banner_video" );
	M_Banner("m_banner_video");

	/*
	** move cursor to a reasonable starting position
	*/
	Menu_AdjustCursor( s_current_menu, 1 );

	/*
	** draw the menu
	*/
	Menu_Draw( s_current_menu );
}

// Knightmare added
void VID_MenuAdvancedDraw (void)
{
	//int w, h;

	s_current_menu = &s_opengl_menu_advanced;

	/*
	** draw the banner
	*/
	//re.DrawGetPicSize( &w, &h, "m_banner_video" );
	//re.DrawPic( viddef.width / 2 - w / 2, viddef.height /2 - 110, "m_banner_video" );
	M_Banner("m_banner_video");

	/*
	** move cursor to a reasonable starting position
	*/
	Menu_AdjustCursor( s_current_menu, 1 );

	/*
	** draw the menu
	*/
	Menu_Draw( s_current_menu );
}

/*
================
VID_MenuKey
================
*/
const char *VID_MenuKey( int key )
{
	menuframework_s *m = s_current_menu;
	char *sound = NULL;

	switch ( key )
	{
	case K_ESCAPE:
		//Knightmare- changed to cancel on exit, thanx to MrG
		//ApplyChanges( 0 );
		CancelChanges(NULL);
		return NULL;
	case K_KP_UPARROW:
	case K_UPARROW:
		m->cursor--;
		Menu_AdjustCursor( m, -1 );
		sound = "misc/menu1.wav";
		break;
	case K_KP_DOWNARROW:
	case K_DOWNARROW:
		m->cursor++;
		Menu_AdjustCursor( m, 1 );
		sound = "misc/menu1.wav";
		break;
	case K_KP_LEFTARROW:
	case K_LEFTARROW:
		Menu_SlideItem( m, -1 );
		sound = "misc/menu1.wav";
		break;
	case K_KP_RIGHTARROW:
	case K_RIGHTARROW:
		Menu_SlideItem( m, 1 );
		sound = "misc/menu1.wav";
		break;
	case K_KP_ENTER:
	case K_ENTER:
		if ( !Menu_SelectItem( m ) )
			ApplyChanges( NULL );
		//sound = "misc/menu1.wav";
		break;
	}

	return sound;
}


